import os
import json
import base64
from functools import wraps
from typing import AsyncIterator, Iterator, Callable, ParamSpec, Awaitable

from pydantic import BaseModel
from openai import OpenAI, AsyncOpenAI, Stream, AsyncStream, APIError as OpenAIAPIError
from openai.types.responses import ResponseStreamEvent

from promptbuilder.llm_client.base_client import BaseLLMClient, BaseLLMClientAsync, ResultType
from promptbuilder.llm_client.types import Response, Content, Candidate, UsageMetadata, Part, ThinkingConfig, Tool, ToolConfig, FunctionCall, MessageDict, Model
from promptbuilder.llm_client.config import DecoratorConfigs
from promptbuilder.llm_client.utils import inherited_decorator
from promptbuilder.llm_client.exceptions import APIError


P = ParamSpec("P")


@inherited_decorator
def _error_handler(func: Callable[P, Response]) -> Callable[P, Response]:
    """
    Decorator to catch error from openai and transform it into unified one
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except OpenAIAPIError as e:
            code = getattr(e, "status_code", None) or e.code
            response = getattr(e, "response", None)
            status = getattr(response, "reason_phrase", None)
            response_json = {
                "status": status,
                "message": e.message,
            }
            raise APIError(code, response_json, response)
    return wrapper

class OpenaiStreamIterator:
    def __init__(self, openai_iterator: Stream[ResponseStreamEvent]):
        self._openai_iterator = openai_iterator

    def __iter__(self) -> Iterator[Response]:
        output_tokens: int | None = None
        input_tokens: int | None = None
        total_tokens: int | None = None
        for next_event in self._openai_iterator:
            if next_event.type == "response.output_text.delta":
                parts = [Part(text=next_event.delta)]
                yield Response(candidates=[Candidate(content=Content(parts=parts, role="model"))])
            elif next_event.type == "response.completed":
                output_tokens = next_event.response.usage.output_tokens
                input_tokens = next_event.response.usage.input_tokens
                total_tokens = next_event.response.usage.total_tokens
        
        usage_metadata = UsageMetadata(
            candidates_token_count=output_tokens,
            prompt_token_count=input_tokens,
            total_token_count=total_tokens,
        )
        yield Response(candidates=[Candidate(content=Content(parts=[Part(text="")], role="model"))], usage_metadata=usage_metadata)


class OpenaiLLMClient(BaseLLMClient):
    PROVIDER = "openai"

    def __init__(
        self,
        model: str,
        api_key: str | None = None,
        decorator_configs: DecoratorConfigs | None = None,
        default_thinking_config: ThinkingConfig | None = None,
        default_max_tokens: int | None = None,
        **kwargs,
    ):
        if api_key is None:
            api_key = os.getenv("OPENAI_API_KEY")
        if api_key is None or not isinstance(api_key, str):
            raise ValueError("To create an openai llm client you need to either set the environment variable OPENAI_API_KEY or pass the api_key in string format")
        super().__init__(OpenaiLLMClient.PROVIDER, model, decorator_configs=decorator_configs, default_thinking_config=default_thinking_config, default_max_tokens=default_max_tokens)
        self._api_key = api_key
        self.client = OpenAI(api_key=api_key)
    
    @property
    def api_key(self) -> str:
        return self._api_key
    
    @staticmethod
    def _content_to_openai_messages(messages: list[Content], system_message: str | None = None) -> list[MessageDict]:
        openai_messages: list[MessageDict] = []
        if system_message is not None:
            openai_messages.append({"role": "developer", "content": system_message})
        for message in messages:
            role = "user" if message.role == "user" else "assistant"
            if message.parts is None:
                openai_messages.append({"role": role, "content": message.as_str()})
            else:
                content = []
                for part in message.parts:
                    if part.inline_data is not None and part.inline_data.data is not None:
                        base64_file_content = base64.b64encode(part.inline_data.data).decode('utf-8')
                        file_data = {
                            "file_data": f"data:{part.inline_data.mime_type};base64,{base64_file_content}"
                        }
                        if part.inline_data.display_name is not None:
                            file_data["filename"] = part.inline_data.display_name
                        match part.inline_data.mime_type:
                            case "application/pdf":
                                file_data["type"] = "input_file"
                            case "image/png" | "image/jpeg" | "image/webp":
                                file_data["type"] = "input_image"
                            case _:
                                raise ValueError(f"Unsupported inline data mime type: {part.inline_data.mime_type}. Supported types are: application/pdf, image/png, image/jpeg, image/webp.")
                        content.append(file_data)
                    else:
                        text_type = "input_text" if message.role == "user" else "output_text"
                        content.append({"type": text_type, "text": part.as_str()})
                openai_messages.append({"role": role, "content": content})
        return openai_messages

    @staticmethod
    def _process_thinking_config(thinking_config: ThinkingConfig | None) -> dict[str, str]:
        if thinking_config is None:
            return {}
        
        openai_thinking_config = {}
        if thinking_config.include_thoughts:
            # openai_thinking_config["summary"] = "auto"
            match thinking_config.thinking_budget:
                case 0 | None:
                    openai_thinking_config["reasoning"] = {"effort": "high"} # default
                case 1:
                    openai_thinking_config["reasoning"] = {"effort": "low"}
                case 2:
                    openai_thinking_config["reasoning"] = {"effort": "medium"}
                case 3:
                    openai_thinking_config["reasoning"] = {"effort": "high"}
                case _:
                    openai_thinking_config["reasoning"] = {"effort": "high"}
        return openai_thinking_config

    @_error_handler
    def _create(
        self,
        messages: list[Content],
        result_type: ResultType = None,
        *,
        thinking_config: ThinkingConfig | None = None,
        system_message: str | None = None,
        max_tokens: int | None = None,
        tools: list[Tool] | None = None,
        tool_config: ToolConfig = ToolConfig(),
    ) -> Response:
        openai_messages: list[MessageDict] = OpenaiLLMClient._content_to_openai_messages(messages, system_message)
        
        if max_tokens is None:
            max_tokens = self.default_max_tokens
        
        openai_kwargs = {
            "model": self.model,
            "max_output_tokens": max_tokens,
            "input": openai_messages,
        }
        
        if thinking_config is None:
            thinking_config = self.default_thinking_config
        openai_kwargs.update(OpenaiLLMClient._process_thinking_config(thinking_config))

        if tools is not None:
            openai_kwargs["parallel_tool_calls"] = True
            
            openai_tools = []
            allowed_function_names = None
            if tool_config.function_calling_config is not None:
                allowed_function_names = tool_config.function_calling_config.allowed_function_names
            for tool in tools:
                for func_decl in tool.function_declarations:
                    if allowed_function_names is None or func_decl.name in allowed_function_names:
                        parameters = func_decl.parameters
                        if parameters is not None:
                            parameters = parameters.model_dump(exclude_none=True)
                            parameters["additionalProperties"] = False
                        else:
                            parameters = {"type": "object", "properties": {}, "required": [], "additionalProperties": False}
                        openai_tools.append({
                            "type": "function",
                            "name": func_decl.name,
                            "description": func_decl.description,
                            "strict": True,
                            "parameters": parameters,
                        })
            openai_kwargs["tools"] = openai_tools
            
            tool_choice_mode = "AUTO"
            if tool_config.function_calling_config is not None:
                if tool_config.function_calling_config.mode is not None:
                    tool_choice_mode = tool_config.function_calling_config.mode
            if tool_choice_mode == "NONE":
                openai_kwargs["tool_choice"] = "none"
            elif tool_choice_mode == "AUTO":
                openai_kwargs["tool_choice"] = "auto"
            elif tool_choice_mode == "ANY":
                openai_kwargs["tool_choice"] = "required"
        
        if result_type is None or result_type == "json":
            response = self.client.responses.create(**openai_kwargs)
            
            parts: list[Part] = []
            for output_item in response.output:
                if output_item.type == "message":
                    for content in output_item.content:
                        parts.append(Part(text=content.text))
                elif output_item.type == "reasoning":
                    for summary in output_item.summary:
                        parts.append(Part(text=summary.text, thought=True))
                elif output_item.type == "function_call":
                    parts.append(Part(function_call=FunctionCall(args=json.loads(output_item.arguments), name=output_item.name)))
            
            return Response(
                candidates=[Candidate(content=Content(parts=parts, role="model"))],
                usage_metadata=UsageMetadata(
                    candidates_token_count=response.usage.output_tokens,
                    prompt_token_count=response.usage.input_tokens,
                    total_token_count=response.usage.total_tokens,
                ),
            )
        elif isinstance(result_type, type(BaseModel)):
            response = self.client.responses.parse(**openai_kwargs, text_format=result_type)
            
            parts: list[Part] = []
            for output_item in response.output:
                if output_item.type == "message":
                    for content in output_item.content:
                        parts.append(Part(text=content.text))
                elif output_item.type == "reasoning":
                    for summary in output_item.summary:
                        parts.append(Part(text=summary.text, thought=True))
                elif output_item.type == "function_call":
                    parts.append(Part(function_call=FunctionCall(args=json.loads(output_item.arguments), name=output_item.name)))
            parsed = response.output_parsed
            
            return Response(
                candidates=[Candidate(content=Content(parts=parts, role="model"))],
                usage_metadata=UsageMetadata(
                    candidates_token_count=response.usage.output_tokens,
                    prompt_token_count=response.usage.input_tokens,
                    total_token_count=response.usage.total_tokens,
                ),
                parsed=parsed,
            )
        else:
            raise ValueError(f"Unsupported result type: {result_type}. Supported types are None, 'json', or a Pydantic model class.")
    
    @_error_handler
    def create_stream(
        self,
        messages: list[Content],
        *,
        thinking_config: ThinkingConfig | None = None,
        system_message: str | None = None,
        max_tokens: int | None = None,
    ) -> Iterator[Response]:
        openai_messages = OpenaiLLMClient._content_to_openai_messages(messages, system_message)
        
        if max_tokens is None:
            max_tokens = self.default_max_tokens
        
        openai_kwargs = {
            "model": self.model,
            "max_output_tokens": max_tokens,
            "input": openai_messages,
        }
        
        if thinking_config is None:
            thinking_config = self.default_thinking_config
        openai_kwargs.update(OpenaiLLMClient._process_thinking_config(thinking_config))
        
        response = self.client.responses.create(**openai_kwargs, stream=True)
        return OpenaiStreamIterator(response)

    @staticmethod
    def models_list() -> list[Model]:
        models: list[Model] = []
        client = OpenAI()
        for openai_model in client.models.list():
            model_name = openai_model.id
            if "tts" in model_name or "whisper" in model_name:
                continue
            if "emb" in model_name:
                continue
            if "davinci" in model_name or "babbage" in model_name:
                continue
            if "image" in model_name or "dall" in model_name:
                continue
            if "moderation" in model_name:
                continue
            if "2024" in model_name or "2025" in model_name:
                continue
            
            models.append(Model(
                full_model_name=OpenaiLLMClient.PROVIDER + ":" + model_name,
                model=model_name,
                provider=OpenaiLLMClient.PROVIDER,
            ))
        return models


@inherited_decorator
def _error_handler_async(func: Callable[P, Awaitable[Response]]) -> Callable[P, Awaitable[Response]]:
    """
    Decorator to catch error from openai and transform it into unified one
    """
    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except OpenAIAPIError as e:
            code = getattr(e, "status_code", None) or e.code
            response = getattr(e, "response", None)
            status = getattr(response, "reason_phrase", None)
            response_json = {
                "status": status,
                "message": e.message,
            }
            raise APIError(code, response_json, response)
    return wrapper

class OpenaiStreamIteratorAsync:
    def __init__(self, openai_iterator: AsyncStream[ResponseStreamEvent]):
        self._openai_iterator = openai_iterator

    async def __aiter__(self) -> AsyncIterator[Response]:
        output_tokens: int | None = None
        input_tokens: int | None = None
        total_tokens: int | None = None
        async for next_event in self._openai_iterator:
            if next_event.type == "response.output_text.delta":
                parts = [Part(text=next_event.delta)]
                yield Response(candidates=[Candidate(content=Content(parts=parts, role="model"))])
            elif next_event.type == "response.completed":
                output_tokens = next_event.response.usage.output_tokens
                input_tokens = next_event.response.usage.input_tokens
                total_tokens = next_event.response.usage.total_tokens
                
        usage_metadata = UsageMetadata(
            candidates_token_count=output_tokens,
            prompt_token_count=input_tokens,
            total_token_count=total_tokens,
        )
        yield Response(candidates=[Candidate(content=Content(parts=[Part(text="")], role="model"))], usage_metadata=usage_metadata)


class OpenaiLLMClientAsync(BaseLLMClientAsync):
    PROVIDER = "openai"
    
    def __init__(
        self,
        model: str,
        api_key: str | None = None,
        decorator_configs: DecoratorConfigs | None = None,
        default_thinking_config: ThinkingConfig | None = None,
        default_max_tokens: int | None = None,
        **kwargs,
    ):
        if api_key is None:
            api_key = os.getenv("OPENAI_API_KEY")
        if api_key is None or not isinstance(api_key, str):
            raise ValueError("To create an openai llm client you need to either set the environment variable OPENAI_API_KEY or pass the api_key in string format")
        super().__init__(OpenaiLLMClientAsync.PROVIDER, model, decorator_configs=decorator_configs, default_thinking_config=default_thinking_config, default_max_tokens=default_max_tokens)
        self._api_key = api_key
        self.client = AsyncOpenAI(api_key=api_key)
    
    @property
    def api_key(self) -> str:
        return self._api_key
    
    @_error_handler_async
    async def _create(
        self,
        messages: list[Content],
        result_type: ResultType = None,
        *,
        thinking_config: ThinkingConfig | None = None,
        system_message: str | None = None,
        max_tokens: int | None = None,
        tools: list[Tool] | None = None,
        tool_config: ToolConfig = ToolConfig(),
    ) -> Response:
        openai_messages = OpenaiLLMClient._content_to_openai_messages(messages, system_message)
        if system_message is not None:
            openai_messages.append({"role": "developer", "content": system_message})
        for message in messages:
            if message.role == "user":
                openai_messages.append({"role": "user", "content": message.as_str()})
            elif message.role == "model":
                openai_messages.append({"role": "assistant", "content": message.as_str()})
        
        if max_tokens is None:
            max_tokens = self.default_max_tokens
        
        openai_kwargs = {
            "model": self.model,
            "max_output_tokens": max_tokens,
            "input": openai_messages,
        }
        
        if thinking_config is None:
            thinking_config = self.default_thinking_config
        openai_kwargs.update(OpenaiLLMClient._process_thinking_config(thinking_config))
        
        if tools is not None:
            openai_kwargs["parallel_tool_calls"] = True
            
            openai_tools = []
            allowed_function_names = None
            if tool_config.function_calling_config is not None:
                allowed_function_names = tool_config.function_calling_config.allowed_function_names
            for tool in tools:
                for func_decl in tool.function_declarations:
                    if allowed_function_names is None or func_decl.name in allowed_function_names:
                        parameters = func_decl.parameters
                        if parameters is not None:
                            parameters = parameters.model_dump(exclude_none=True)
                            parameters["additionalProperties"] = False
                        else:
                            parameters = {"type": "object", "properties": {}, "required": [], "additionalProperties": False}
                        openai_tools.append({
                            "type": "function",
                            "name": func_decl.name,
                            "description": func_decl.description,
                            "strict": True,
                            "parameters": parameters,
                        })
            openai_kwargs["tools"] = openai_tools
            
            tool_choice_mode = "AUTO"
            if tool_config.function_calling_config is not None:
                if tool_config.function_calling_config.mode is not None:
                    tool_choice_mode = tool_config.function_calling_config.mode
            if tool_choice_mode == "NONE":
                openai_kwargs["tool_choice"] = "none"
            elif tool_choice_mode == "AUTO":
                openai_kwargs["tool_choice"] = "auto"
            elif tool_choice_mode == "ANY":
                openai_kwargs["tool_choice"] = "required"
        
        if result_type is None or result_type == "json":
            response = await self.client.responses.create(**openai_kwargs)
            
            parts: list[Part] = []
            for output_item in response.output:
                if output_item.type == "message":
                    for content in output_item.content:
                        parts.append(Part(text=content.text))
                elif output_item.type == "reasoning":
                    for summary in output_item.summary:
                        parts.append(Part(text=summary.text, thought=True))
                elif output_item.type == "function_call":
                    parts.append(Part(function_call=FunctionCall(args=json.loads(output_item.arguments), name=output_item.name)))
            
            return Response(
                candidates=[Candidate(content=Content(parts=parts, role="model"))],
                usage_metadata=UsageMetadata(
                    candidates_token_count=response.usage.output_tokens,
                    prompt_token_count=response.usage.input_tokens,
                    total_token_count=response.usage.total_tokens,
                ),
            )
        elif isinstance(result_type, type(BaseModel)):
            response = await self.client.responses.parse(**openai_kwargs, text_format=result_type)
            
            parts: list[Part] = []
            for output_item in response.output:
                if output_item.type == "message":
                    for content in output_item.content:
                        parts.append(Part(text=content.text))
                elif output_item.type == "reasoning":
                    for summary in output_item.summary:
                        parts.append(Part(text=summary.text, thought=True))
                elif output_item.type == "function_call":
                    parts.append(Part(function_call=FunctionCall(args=json.loads(output_item.arguments), name=output_item.name)))
            parsed = response.output_parsed
            
            return Response(
                candidates=[Candidate(content=Content(parts=parts, role="model"))],
                usage_metadata=UsageMetadata(
                    candidates_token_count=response.usage.output_tokens,
                    prompt_token_count=response.usage.input_tokens,
                    total_token_count=response.usage.total_tokens,
                ),
                parsed=parsed,
            )
        else:
            raise ValueError(f"Unsupported result_type: {result_type}. Supported types are: None, 'json', or a Pydantic model.")
    
    @_error_handler_async
    async def create_stream(
        self,
        messages: list[Content],
        *,
        thinking_config: ThinkingConfig | None = None,
        system_message: str | None = None,
        max_tokens: int | None = None,
    ) -> AsyncIterator[Response]:
        openai_messages = OpenaiLLMClient._content_to_openai_messages(messages, system_message)
        if system_message is not None:
            openai_messages.append({"role": "developer", "content": system_message})
        for message in messages:
            if message.role == "user":
                openai_messages.append({"role": "user", "content": message.as_str()})
            elif message.role == "model":
                openai_messages.append({"role": "assistant", "content": message.as_str()})
        
        if max_tokens is None:
            max_tokens = self.default_max_tokens
        
        openai_kwargs = {
            "model": self.model,
            "max_output_tokens": max_tokens,
            "input": openai_messages,
        }
        
        if thinking_config is None:
            thinking_config = self.default_thinking_config
        openai_kwargs.update(OpenaiLLMClient._process_thinking_config(thinking_config))
        
        response = await self.client.responses.create(**openai_kwargs, stream=True)
        return OpenaiStreamIteratorAsync(response)

    @staticmethod
    def models_list() -> list[Model]:
        return OpenaiLLMClient.models_list()



from typing import Callable, ParamSpec, TypeVar, Awaitable, Generic, Type, Any

from pydantic import BaseModel

from promptbuilder.llm_client.types import Tool, FunctionDeclaration, Schema


P = ParamSpec("P")
T = TypeVar("T")

class CallableTool(BaseModel, Generic[P, T]):
    arg_descriptions: dict[str, str] = {}
    function: Callable[P, Awaitable[T]]

    model_config = {"extra": "allow"}
    
    def model_post_init(self, __context: Any):
        self.name = self.function.__name__
        self.tool = self._make_tool()
        return super().model_post_init(__context)

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return await self.function(*args, **kwargs)

    @staticmethod
    def description_without_indent(description: str) -> str:
        lines = description.strip().splitlines()
        indent = min(len(line) - len(line.lstrip()) for line in lines if line.strip())
        return "\n".join(line[indent:] for line in lines)

    @staticmethod
    def _type_to_str(type: Type) -> str:
        if type is str:
            return "string"
        elif type is float:
            return "number"
        elif type is int:
            return "integer"
        elif type is bool:
            return "boolean"
        elif type is list:
            return "array"
        elif type is dict:
            return "object"
        else:
            raise ValueError("Unsupported argument type")
    
    def _make_tool(self) -> Tool:
        tool_name = self.function.__name__
        args = {name: type for name, type in self.function.__annotations__.items() if name != "return"}
        description = CallableTool.description_without_indent(self.function.__doc__)

        return Tool(
            function_declarations=[
                FunctionDeclaration(
                    name=tool_name,
                    description=description,
                    parameters=(
                        Schema(
                            type=self._type_to_str(dict),
                            properties={
                                name: Schema(
                                    type=self._type_to_str(type),
                                    description=self.arg_descriptions.get(name, None),
                                )
                                for name, type in args.items()
                            },
                        )
                        if len(args) > 0
                        else None
                    )
                )
            ],
            callable=self.function
        )




import json
from typing import Any

import httpx
import aiohttp


class APIError(Exception):
    """General errors raised by the llm clients"""
    code: int | None = None
    status: str | None = None
    message: str | None = None
    response: httpx.Response | aiohttp.ClientResponse | None = None

    def __init__(self, code: int | None = None, response_json: dict[str, Any] = {}, response: httpx.Response | aiohttp.ClientResponse | None = None):
        self.response = response
        self.details = response_json
        self.message = self._get_message(response_json)
        self.status = self._get_status(response_json)
        self.code = code if code else self._get_code(response_json)

        super().__init__(f"{self.code} {self.status}. {self.details}")

    def _get_status(self, response_json: dict[str, Any]) -> Any:
        return response_json.get(
            "status", response_json.get("error", {}).get("status", None)
        )

    def _get_message(self, response_json: dict[str, Any]) -> Any:
        return response_json.get(
            "message", response_json.get("error", {}).get("message", None)
        )

    def _get_code(self, response_json: dict[str, Any]) -> Any:
        return response_json.get(
            "code", response_json.get("error", {}).get("code", None)
        )

    def _to_replay_record(self) -> dict[str, Any]:
        """Returns a dictionary representation of the error for replay recording."""
        return {
            "error": {
                "code": self.code,
                "message": self.message,
                "status": self.status,
            }
        }

    @classmethod
    def raise_for_response(cls, response: httpx.Response):
        """Raises an error with detailed error message if the response has an error status."""
        if response.status_code == 200:
            return

        if isinstance(response, httpx.Response):
            try:
                response.read()
                response_json = response.json()
            except json.decoder.JSONDecodeError:
                message = response.text
                response_json = {
                    "message": message,
                    "status": response.reason_phrase,
                }
        else:
            response_json = response.body_segments[0].get('error', {})

        status_code = response.status_code
        if 400 <= status_code < 500:
            raise ClientError(status_code, response_json, response)
        elif 500 <= status_code < 600:
            raise ServerError(status_code, response_json, response)
        else:
            raise cls(status_code, response_json, response)

    @classmethod
    async def raise_for_async_response(cls, response: httpx.Response | aiohttp.ClientResponse):
        """Raises an error with detailed error message if the response has an error status."""
        status_code = 0
        response_json = None
        if isinstance(response, httpx.Response):
            if response.status_code == 200:
                return
            try:
                await response.aread()
                response_json = response.json()
            except json.decoder.JSONDecodeError:
                message = response.text
                response_json = {
                    "message": message,
                    "status": response.reason_phrase,
                }
            status_code = response.status_code
        else:
            try:
                if isinstance(response, aiohttp.ClientResponse):
                    if response.status == 200:
                        return
                    try:
                        response_json = await response.json()
                    except aiohttp.client_exceptions.ContentTypeError:
                        message = await response.text()
                        response_json = {
                            "message": message,
                            "status": response.reason,
                        }
                    status_code = response.status
                else:
                    response_json = response.body_segments[0].get("error", {})
            except ImportError:
                response_json = response.body_segments[0].get("error", {})

        if 400 <= status_code < 500:
            raise ClientError(status_code, response_json, response)
        elif 500 <= status_code < 600:
            raise ServerError(status_code, response_json, response)
        else:
            raise cls(status_code, response_json, response)


class ClientError(APIError):
    """Client error raised by the llm clients"""
    pass


class ServerError(APIError):
    """Server error raised by the llm clients"""
    pass



import inspect
from typing import Callable, ParamSpec, TypeVar, Awaitable


P = ParamSpec("P")
T = TypeVar("T")

async def run_async(f: Callable[P, Awaitable[T]] | Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
    if inspect.iscoroutinefunction(f):
        return await f(*args, **kwargs)
    else:
        return f(*args, **kwargs)



from setuptools import setup, find_packages

setup(
    name="promptbuilder",
    version="0.4.25",
    packages=find_packages(),
    install_requires=[
        "pydantic",
        "pytest",
        "aisuite-async",
        "google-genai>=1.4.0",
        "anthropic",
        "openai",
        "aioboto3"
    ],
    author="Kapulkin Stanislav",
    author_email="kapulkin@gmail.com",
    description="Library for building prompts for LLMs",
    long_description=open("Readme.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/kapulkin/promptbuilder",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)


import os
from functools import wraps
from typing import AsyncIterator, Iterator, Callable, ParamSpec, Awaitable

from pydantic import BaseModel
from tenacity import RetryError
from google.genai import Client, types

from promptbuilder.llm_client.base_client import BaseLLMClient, BaseLLMClientAsync, ResultType
from promptbuilder.llm_client.types import Response, Content, Part, ThinkingConfig, Tool, ToolConfig, Model
from promptbuilder.llm_client.config import DecoratorConfigs
from promptbuilder.llm_client.utils import inherited_decorator
from promptbuilder.llm_client.exceptions import APIError


P = ParamSpec("P")


@inherited_decorator
def _error_handler(func: Callable[P, Response]) -> Callable[P, Response]:
    """
    Decorator to catch error from google.genai and transform it into unified one
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except RetryError as retry_error:
            e = retry_error.last_attempt._exception
            if e is None:
                raise APIError()
            code = e.code
            response_json = {
                "status": e.status,
                "message": e.message,
            }
            response = e.response
            raise APIError(code, response_json, response)
    return wrapper


class GoogleLLMClient(BaseLLMClient):
    PROVIDER: str = "google"
    
    def __init__(
        self,
        model: str,
        api_key: str | None = None,
        decorator_configs: DecoratorConfigs | None = None,
        default_thinking_config: ThinkingConfig | None = None,
        default_max_tokens: int | None = None,
        **kwargs,
    ):
        if api_key is None:
            api_key = os.getenv("GOOGLE_API_KEY")
        if api_key is None or not isinstance(api_key, str):
            raise ValueError("To create a google llm client you need to either set the environment variable GOOGLE_API_KEY or pass the api_key in string format")
        super().__init__(GoogleLLMClient.PROVIDER, model, decorator_configs=decorator_configs, default_thinking_config=default_thinking_config, default_max_tokens=default_max_tokens)
        self._api_key = api_key
        self.client = Client(api_key=api_key)
    
    @property
    def api_key(self) -> str:
        return self._api_key
    
    @staticmethod
    def _preprocess_messages(messages: list[Content]) -> list[Content]:
        new_messages = []
        for message in messages:
            # TODO:
            # copy parts from message to new_message
            # if part has inline_data, set display_name to None in new_message
            new_parts = []
            if message.parts:
                for part in message.parts:
                    if part.inline_data is not None:
                        new_part = Part.model_copy(part, deep=True)
                        new_part.inline_data.display_name = None
                    else:
                        new_part = part
                    new_parts.append(new_part)
            new_message = Content(
                role=message.role,
                parts=new_parts
            )
            new_messages.append(new_message)
        return new_messages
    
    @_error_handler
    def _create(
        self,
        messages: list[Content],
        result_type: ResultType = None,
        *,
        thinking_config: ThinkingConfig | None = None,
        system_message: str | None = None,
        max_tokens: int | None = None,
        tools: list[Tool] | None = None,
        tool_config: ToolConfig = ToolConfig(),
    ) -> Response:
        messages = self._preprocess_messages(messages)
        if max_tokens is None:
            max_tokens = self.default_max_tokens
        config = types.GenerateContentConfig(
            system_instruction=system_message,
            max_output_tokens=max_tokens,
            tools=tools,
            tool_config=tool_config,
        )
        
        if thinking_config is None:
            thinking_config = self.default_thinking_config
        config.thinking_config = thinking_config
        
        if result_type is None:
            return self.client.models.generate_content(
                model=self.model,
                contents=messages,
                config=config,
            )
        elif result_type == "json":
            response = self.client.models.generate_content(
                model=self.model,
                contents=messages,
                config=config,
            )
            response.parsed = BaseLLMClient.as_json(response.text)
            return response
        elif isinstance(result_type, type(BaseModel)):
            config.response_mime_type = "application/json"
            config.response_schema = result_type
            return self.client.models.generate_content(
                model=self.model,
                contents=messages,
                config=config,
            )
        else:
            raise ValueError(f"Unsupported result_type: {result_type}. Supported types are: None, 'json', or a Pydantic model.")
    
    @_error_handler
    def create_stream(
        self,
        messages: list[Content],
        *,
        thinking_config: ThinkingConfig | None = None,
        system_message: str | None = None,
        max_tokens: int | None = None,
    ) -> Iterator[Response]:
        if max_tokens is None:
            max_tokens = self.default_max_tokens
        config = types.GenerateContentConfig(
            system_instruction=system_message,
            max_output_tokens=max_tokens,
        )
        
        if thinking_config is None:
            thinking_config = self.default_thinking_config
        config.thinking_config = thinking_config
        
        response = self.client.models.generate_content_stream(
            model=self.model,
            contents=messages,
            config=config,
        )
        return response
    
    @staticmethod
    def models_list() -> list[Model]:
        models: list[Model] = []
        client = Client()
        for google_model in client.models.list():
            for action in google_model.supported_actions:
                if action == "generateContent":
                    model_name = google_model.name
                    if model_name.startswith("models/"):
                        model_name = model_name[7:]
                    
                    if "tts" in model_name.lower():
                        continue
                    if "emb" in model_name.lower():
                        continue
                    if "image-generation" in model_name.lower():
                        continue
                    if "gemini" not in model_name.lower():
                        continue
                    
                    models.append(Model(
                        full_model_name=GoogleLLMClient.PROVIDER + ":" + model_name,
                        model=model_name,
                        provider=GoogleLLMClient.PROVIDER,
                        display_name=google_model.display_name,
                    ))
        return models


@inherited_decorator
def _error_handler_async(func: Callable[P, Awaitable[Response]]) -> Callable[P, Awaitable[Response]]:
    """
    Decorator to catch error from google.genai and transform it into unified one
    """
    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except RetryError as retry_error:
            e = retry_error.last_attempt._exception
            if e is None:
                raise APIError()
            code = e.code
            response_json = {
                "status": e.status,
                "message": e.message,
            }
            response = e.response
            raise APIError(code, response_json, response)
    return wrapper

class GoogleLLMClientAsync(BaseLLMClientAsync):
    PROVIDER: str = "google"
    
    def __init__(
        self,
        model: str,
        api_key: str | None = None,
        decorator_configs: DecoratorConfigs | None = None,
        default_thinking_config: ThinkingConfig | None = None,
        default_max_tokens: int | None = None,
        **kwargs,
    ):
        if api_key is None:
            api_key = os.getenv("GOOGLE_API_KEY")
        if api_key is None or not isinstance(api_key, str):
            raise ValueError("To create a google llm client you need to either set the environment variable GOOGLE_API_KEY or pass the api_key in string format")
        super().__init__(GoogleLLMClientAsync.PROVIDER, model, decorator_configs=decorator_configs, default_thinking_config=default_thinking_config, default_max_tokens=default_max_tokens)
        self._api_key = api_key
        self.client = Client(api_key=api_key)
    
    @property
    def api_key(self) -> str:
        return self._api_key

    @_error_handler_async
    async def _create(
        self,
        messages: list[Content],
        result_type: ResultType = None,
        *,
        thinking_config: ThinkingConfig | None = None,
        system_message: str | None = None,
        max_tokens: int | None = None,
        tools: list[Tool] | None = None,
        tool_config: ToolConfig = ToolConfig(),
    ) -> Response:
        messages = GoogleLLMClient._preprocess_messages(messages)
        if max_tokens is None:
            max_tokens = self.default_max_tokens
        config = types.GenerateContentConfig(
            system_instruction=system_message,
            max_output_tokens=max_tokens,
            tools=tools,
            tool_config=tool_config,
        )
        
        if thinking_config is None:
            thinking_config = self.default_thinking_config
        config.thinking_config = thinking_config
        
        if result_type is None or result_type == "json":
            return await self.client.aio.models.generate_content(
                model=self.model,
                contents=messages,
                config=config,
            )
        elif isinstance(result_type, type(BaseModel)):
            config.response_mime_type = "application/json"
            config.response_schema = result_type
            return await self.client.aio.models.generate_content(
                model=self.model,
                contents=messages,
                config=config,
            )
        else:
            raise ValueError(f"Unsupported result_type: {result_type}. Supported types are: None, 'json', or a Pydantic model.")
    
    @_error_handler_async
    async def create_stream(
        self,
        messages: list[Content],
        *,
        thinking_config: ThinkingConfig | None = None,
        system_message: str | None = None,
        max_tokens: int | None = None,
    ) -> AsyncIterator[Response]:
        if max_tokens is None:
            max_tokens = self.default_max_tokens
        config = types.GenerateContentConfig(
            system_instruction=system_message,
            max_output_tokens=max_tokens,
        )
        
        if thinking_config is None:
            thinking_config = self.default_thinking_config
        config.thinking_config = thinking_config
        
        response = await self.client.aio.models.generate_content_stream(
            model=self.model,
            contents=messages,
            config=config,
        )
        return response

    @staticmethod
    def models_list() -> list[Model]:
        return GoogleLLMClient.models_list()



from typing import Type, Union, Literal, get_origin, get_args, Self
from types import UnionType
from pydantic import Field, BaseModel
from pydantic_core import PydanticUndefined
from enum import Enum

class Prompt:
    def __init__(self, prompt_template: str, variables: list[str]):
        self.prompt_template = prompt_template
        self.variables = variables
    
    def render(self, **kwargs) -> str:
        return self.prompt_template.format(**{v: kwargs[v] for v in self.variables})

    def format(self, **kwargs) -> str:
        return self.render(**kwargs)

    def __str__(self) -> str:
        return self.render()

class PromptBuilder:
    def __init__(self):
        self.prompt_template = ""
        self.variables = []

    def tag_variable(self, tag_name: str, variable_name: str, description: str | None = None) -> Self:
        if description:
            self.prompt_template += f"{description}\n"
        self.prompt_template += f"<{tag_name}>\n{{{variable_name}}}\n</{tag_name}>\n"
        self.variables.append(variable_name)
        return self

    def tag_content(self, tag_name: str, content: str, description: str | None = None) -> Self:
        if description:
            self.prompt_template += f"{description}\n"
        self.prompt_template += f"<{tag_name}>\n{content}\n</{tag_name}>\n"
        return self
    
    def text(self, text: str) -> Self:
        self.prompt_template += text
        return self

    def header(self, header: str, level: int = 2) -> Self:
        self.prompt_template += f"\n{'#' * level} {header}\n"
        return self

    def variable(self, variable_name: str) -> Self:
        self.prompt_template += f"{{{variable_name}}}"
        self.variables.append(variable_name)
        return self

    def paragraph(self, text: str) -> Self:
        self.prompt_template += f"{text}\n"
        return self

    def structure(self, type: Type, description: str | None = None, rebuild_models: bool = False) -> Self:
        if description:
            self.prompt_template += f"{description}\n"
        ts_type = schema_to_ts(type, rebuild_models=rebuild_models)

        circular_type_printed = False
        for index, (name, dep_type) in enumerate(ts_type.circular_types.items()):
            if name != ts_type.name:
                if index > 0:
                    self.prompt_template += "\n\n"
                self.prompt_template += f"{name}: {dep_type}"
                circular_type_printed = True
        
        if circular_type_printed > 0:
            self.prompt_template += "\n\n"
        if ts_type.name in ts_type.circular_types:
            self.prompt_template += f"{ts_type.name}: "
        self.prompt_template += ts_type.type
        return self
    
    def set_structured_output(self, type: Type, output_name: str = "result", rebuild_models: bool = False) -> Self:
        """
        Set structured output for the prompt.

        Use create_model from pydantic to define the structure on the fly.

        Example:
        builder.set_structured_output(type=create_model(
            "TodoList",
            todo_items=(List[create_model(
                "TodoItem",
                description=(str, Field()),
                is_done=(bool, Field())
            )], Field())
        ))
        """
        self.prompt_template += f"Return {output_name} in a following JSON structure:\n"
        self.structure(type, rebuild_models=rebuild_models)
        self.prompt_template += "\nYour output should consist solely of the JSON object, with no additional text."
        return self

    def build(self) -> Prompt:
        return Prompt(self.prompt_template, self.variables)

class TypeScriptType(BaseModel):
    pass

class TypeScriptTypeWithDependencies(BaseModel):
    name: str | None = None
    type: str
    dependencies: set[str] = set()
    circular_types: dict[str, str] = {}

def _uinoin_types(ts_types: list[TypeScriptTypeWithDependencies]) -> TypeScriptTypeWithDependencies:
    type = ' | '.join([ts.type for ts in ts_types])
    dependencies = set[str]()
    circular_types = {}
    for ts in ts_types:
        dependencies.update(ts.dependencies)
        circular_types.update(ts.circular_types)
    return TypeScriptTypeWithDependencies(type=type, dependencies=dependencies, circular_types=circular_types)

def _type_with_fields(type_name: str, fields, indent_str: str) -> str:
    return '{{\n' + '\n'.join(fields) + f'\n{indent_str}' + '}}' if len(fields) > 0 else f"'{type_name}'"

def _unindent(text: str, indent: int) -> str:
    """Remove leading indentation from each line in the text."""
    lines = text.splitlines()
    unindented_lines = [line[indent:] if len(line) >= indent and all(c == ' ' for c in line[:indent]) else line for line in lines]
    return '\n'.join(unindented_lines)

def _schema_to_ts(value_type, parent_types: set[str], indent: int = 2, depth: int = 0, rebuild_models: bool = False) -> TypeScriptTypeWithDependencies:
    """Convert Pydantic model to TypeScript type notation string."""

    # Handle basic types directly
    if value_type == str:
        return TypeScriptTypeWithDependencies(type='string')
    if value_type in (int, float):
        return TypeScriptTypeWithDependencies(type='number')
    if value_type == bool:
        return TypeScriptTypeWithDependencies(type='boolean')
    if value_type == type(None):
        return TypeScriptTypeWithDependencies(type='null')
    if value_type == None:
        return TypeScriptTypeWithDependencies(type='any')
    
    origin = get_origin(value_type)
    # Handle Literal types
    if origin == Literal:
        literal_args = get_args(value_type)
        # Convert each literal value to a string representation
        literal_values = []
        for arg in literal_args:
            if isinstance(arg, str):
                literal_values.append(TypeScriptTypeWithDependencies(type=f"'{arg}'"))
            elif isinstance(arg, (int, float, bool)):
                literal_values.append(TypeScriptTypeWithDependencies(type=str(arg)))
            else:
                literal_values.append(_schema_to_ts(arg, parent_types, indent, depth, rebuild_models))
        return _uinoin_types(literal_values)
    
    # Handle Enum types
    if isinstance(value_type, type) and issubclass(value_type, Enum):
        # Convert enum values to TypeScript union type
        enum_values = []
        for v in value_type:
            if isinstance(v.value, str):
                enum_values.append(f"'{v.value}'")
            elif isinstance(v.value, (int, float, bool)):
                enum_values.append(str(v.value))
            else:
                enum_values.append(f"'{str(v.value)}'")
        return TypeScriptTypeWithDependencies(type=' | '.join(enum_values))
    
    # Handle list types
    if origin == list:
        list_type_args = get_args(value_type)
        if list_type_args:
            list_type_arg = list_type_args[0]
            is_multi_union = get_origin(list_type_arg) == UnionType and len(get_args(list_type_arg)) > 1
            item_ts_type = _schema_to_ts(list_type_arg, parent_types, indent, depth, rebuild_models)
            item_ts_type_str = f"({item_ts_type.type})" if is_multi_union else item_ts_type.type
            return TypeScriptTypeWithDependencies(type=f'{item_ts_type_str}[]', dependencies=item_ts_type.dependencies, circular_types=item_ts_type.circular_types)
        return TypeScriptTypeWithDependencies(type='any[]')
    
    # Handle dict types
    if origin == dict:
        dict_type_args = get_args(value_type)
        if len(dict_type_args) == 2:
            key_type = _schema_to_ts(dict_type_args[0], parent_types, indent, depth, rebuild_models)
            value_type = _schema_to_ts(dict_type_args[1], parent_types, indent, depth, rebuild_models)
            # In TypeScript, only string, number, and symbol can be used as index types
            if key_type.type not in ['string', 'number']:
                key_type.type = 'string'
            dependencies = key_type.dependencies.union(value_type.dependencies)
            circular_types = key_type.circular_types.copy()
            circular_types.update(value_type.circular_types)
            return TypeScriptTypeWithDependencies(
                type='{{' + f' [key: {key_type.type}]: {value_type.type}' + '}}',
                dependencies=dependencies, circular_types=circular_types
            )
        return TypeScriptTypeWithDependencies(type='{{ [key: string]: any }}')
    
    # Handle Union types
    if origin == UnionType or origin == Union:
        union_args = get_args(value_type)

        arg_class_type_names = set[str]()
        for arg in union_args:
            if hasattr(arg, 'model_fields'):
                arg_class_type_names.add(arg.__name__)
            # put into tt parent_types united with other_types except arg.__name__

        ts_types = []
        for arg in union_args:
            if hasattr(arg, 'model_fields'):
                other_type_names = arg_class_type_names - {arg.__name__}
            else:
                other_type_names = arg_class_type_names

            other_type_names = other_type_names.union(parent_types)
            ts_types.append(_schema_to_ts(arg, other_type_names, indent, depth, rebuild_models))
        
        name_to_ts_types = {ts.name: ts for ts in ts_types if ts.name is not None}

        union_ts_type = _uinoin_types(ts_types)
        for name in arg_class_type_names:
            if name in union_ts_type.dependencies:
                circular_type = name_to_ts_types[name]
                if circular_type.name != circular_type.type:
                    union_ts_type.circular_types[name] = _unindent(circular_type.type, indent * depth)
        return union_ts_type

    # If not a Pydantic model, return any
    if not hasattr(value_type, 'model_fields'):
        return TypeScriptTypeWithDependencies(type='any')
    
    # Handle Pydantic models
    indent_str = ' ' * indent * depth
    fields_indent_str = indent_str + ' ' * indent
    fields = []

    # Check for circular dependencies
    type_name = value_type.__name__
    if type_name in parent_types:
        return TypeScriptTypeWithDependencies(name=type_name, type=type_name, dependencies={ type_name })

    parent_types = parent_types.union({type_name})
    dependencies = set[str]()
    circular_types = {}

    if rebuild_models:
        value_type.model_rebuild()

    for field_name, field in value_type.model_fields.items():
        field_type = field.annotation
        if field_type == TypeScriptType:
            ts_type = TypeScriptTypeWithDependencies(type=field.title)
        else:
            ts_type = _schema_to_ts(field_type, parent_types, indent, depth + 1, rebuild_models)
            dependencies.update(ts_type.dependencies)
            circular_types.update(ts_type.circular_types)
            
        # Add question mark for optional fields (those with default values or None default)
        is_optional = (field.default is not None and field.default is not PydanticUndefined) or field.default_factory is not None
        optional_marker = '?' if is_optional else ''
            
        # Add field description if available
        description = field.description or ''
        if description:
            fields.append(f'{fields_indent_str}{field_name}{optional_marker}: {ts_type.type}, // {description}')
        else:
            fields.append(f'{fields_indent_str}{field_name}{optional_marker}: {ts_type.type},')

    type_str = _type_with_fields(type_name, fields, indent_str)

    if type_name in dependencies:
        circular_types[type_name] = _unindent(type_str, indent * depth)
        if depth > 0:
            return TypeScriptTypeWithDependencies(name=type_name, type=type_name, dependencies=dependencies, circular_types=circular_types)

    return TypeScriptTypeWithDependencies(name=type_name, type=type_str, dependencies=dependencies, circular_types=circular_types)

def schema_to_ts(value_type, indent: int = 2, rebuild_models: bool = False) -> TypeScriptTypeWithDependencies:
    return _schema_to_ts(value_type, set[str](), indent, 0, rebuild_models)



import logging
from abc import ABC, abstractmethod
from typing import Optional, Any, Callable, Literal, TypeVar, Self, Protocol, runtime_checkable
from enum import Enum

from pydantic import BaseModel, model_validator, ConfigDict


logger = logging.getLogger(__name__)

type MessagesDict = list[dict[str, str]]
type Role = Literal["user", "model"]
type Json = list | dict
type JsonType = Literal["string", "number", "integer", "boolean", "array", "object"]
PydanticStructure = TypeVar("PydanticStructure", bound=BaseModel)

type ResultType = Literal["json"] | type[PydanticStructure] | None


@runtime_checkable
class PartLike(Protocol):
    """Protocol for Part-like objects that have the same interface as Part."""
    text: Optional[str]
    function_call: Optional[Any]  # Using Any to allow different FunctionCall types
    function_response: Optional[Any]  # Using Any to allow different FunctionResponse types
    thought: Optional[bool]
    inline_data: Optional[Any]  # Using Any to allow different Blob types


class CustomApiKey(ABC):
    @abstractmethod
    def __hash__(self):
        pass

type ApiKey = str | CustomApiKey

type MessageDict = dict[str, str | Any | None]

class Message(BaseModel):
    role: str
    content: str | dict[str, Any] | None = None

class Choice(BaseModel):
    message: Message

class Usage(BaseModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int

class Completion(BaseModel):
    choices: list[Choice]
    usage: Optional[Usage] = None

class FunctionCall(BaseModel):
    id: Optional[str] = None
    args: Optional[dict[str, Any]] = None
    name: Optional[str] = None

class FunctionResponse(BaseModel):
    id: Optional[str] = None
    name: Optional[str] = None
    response: Optional[dict[str, Any]] = None

class Blob(BaseModel):
    display_name: Optional[str] = None
    data: Optional[bytes] = None
    mime_type: Optional[str] = None


class Part(BaseModel):
    text: Optional[str] = None
    function_call: Optional[FunctionCall] = None
    function_response: Optional[FunctionResponse] = None
    thought: Optional[bool] = None
    inline_data: Optional[Blob] = None
    
    def as_str(self) -> str:
        if self.text is not None:
            return self.text
        return ""

    @classmethod
    def from_bytes(cls, *, data: bytes, mime_type: str, display_name: str | None = None) -> 'Part':
        inline_data = Blob(
            data=data,
            mime_type=mime_type,
            display_name=display_name
        )
        return cls(inline_data=inline_data)

class Content(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    
    parts: list[Part | PartLike] | None = None
    role: Role | None = None
    
    def as_str(self) -> str:
        if self.parts is None:
            return ""
        else:
            return "\n".join([(part.text or "") for part in self.parts])


class FinishReason(Enum):
  """Output only. The reason why the model stopped generating tokens.

  If empty, the model has not stopped generating the tokens.
  """

  FINISH_REASON_UNSPECIFIED = 'FINISH_REASON_UNSPECIFIED'
  """The finish reason is unspecified."""
  STOP = 'STOP'
  """Token generation reached a natural stopping point or a configured stop sequence."""
  MAX_TOKENS = 'MAX_TOKENS'
  """Token generation reached the configured maximum output tokens."""
  SAFETY = 'SAFETY'
  """Token generation stopped because the content potentially contains safety violations. NOTE: When streaming, [content][] is empty if content filters blocks the output."""
  RECITATION = 'RECITATION'
  """The token generation stopped because of potential recitation."""
  LANGUAGE = 'LANGUAGE'
  """The token generation stopped because of using an unsupported language."""
  OTHER = 'OTHER'
  """All other reasons that stopped the token generation."""
  BLOCKLIST = 'BLOCKLIST'
  """Token generation stopped because the content contains forbidden terms."""
  PROHIBITED_CONTENT = 'PROHIBITED_CONTENT'
  """Token generation stopped for potentially containing prohibited content."""
  SPII = 'SPII'
  """Token generation stopped because the content potentially contains Sensitive Personally Identifiable Information (SPII)."""
  MALFORMED_FUNCTION_CALL = 'MALFORMED_FUNCTION_CALL'
  """The function call generated by the model is invalid."""
  IMAGE_SAFETY = 'IMAGE_SAFETY'
  """Token generation stopped because generated images have safety violations."""
  UNEXPECTED_TOOL_CALL = 'UNEXPECTED_TOOL_CALL'
  """The tool call generated by the model is invalid."""

class Candidate(BaseModel):
    content: Optional[Content] = None
    finish_reason: Optional[FinishReason] = None

class UsageMetadata(BaseModel):
    cached_content_token_count: Optional[int] = None
    candidates_token_count: Optional[int] = None
    prompt_token_count: Optional[int] = None
    thoughts_token_count: Optional[int] = None
    total_token_count: Optional[int] = None

class ThinkingConfig(BaseModel):
    include_thoughts: Optional[bool] = None
    thinking_budget: Optional[int] = None

class Response(BaseModel):
    candidates: Optional[list[Candidate]] = None
    usage_metadata: Optional[UsageMetadata] = None
    parsed: Optional[Json | PydanticStructure] = None

    @property
    def text(self) -> Optional[str]:
        """Returns the concatenation of all text parts in the response."""
        if (
            not self.candidates
            or not self.candidates[0].content
            or not self.candidates[0].content.parts
        ):
            return None
        if len(self.candidates) > 1:
            logger.warning(
                f"there are {len(self.candidates)} candidates, returning text from"
                " the first candidate.Access response.candidates directly to get"
                " text from other candidates."
            )
        text = ""
        any_text_part_text = False
        for part in self.candidates[0].content.parts:
            if isinstance(part.text, str):
                if isinstance(part.thought, bool) and part.thought:
                    continue
                any_text_part_text = True
                text += part.text
        # part.text == "" is different from part.text is None
        return text if any_text_part_text else None

class Schema(BaseModel):
    example: Optional[Any] = None
    pattern: Optional[str] = None
    minimum: Optional[float] = None
    default: Optional[Any] = None
    any_of: Optional[list["Schema"]] = None
    max_length: Optional[int] = None
    title: Optional[str] = None
    min_length: Optional[int] = None
    min_properties: Optional[int] = None
    maximum: Optional[float] = None
    max_properties: Optional[int] = None
    description: Optional[str] = None
    enum: Optional[list[str]] = None
    format: Optional[str] = None
    items: Optional["Schema"] = None
    max_items: Optional[int] = None
    min_items: Optional[int] = None
    nullable: Optional[bool] = None
    properties: Optional[dict[str, "Schema"]] = None
    property_ordering: Optional[list[str]] = None
    required: Optional[list[str]] = None
    type: Optional[JsonType] = None

class FunctionDeclaration(BaseModel):
    response: Optional[Schema] = None
    description: Optional[str] = None
    name: Optional[str] = None
    parameters: Optional[Schema] = None

class Tool(BaseModel):
    function_declarations: Optional[list[FunctionDeclaration]] = None
    callable: Optional[Callable] = None

class FunctionCallingConfig(BaseModel):
    mode: Optional[Literal["AUTO", "ANY", "NONE"]] = None
    allowed_function_names: Optional[list[str]] = None

class ToolConfig(BaseModel):
    function_calling_config: Optional[FunctionCallingConfig] = None

class Model(BaseModel):
    full_model_name: str
    provider: str
    model: str
    display_name: str | None = None
